
;; generic functional combinators

; fast ones (no apply)

(define (id x) x) ; I
(define (dot f g) (lambda (x) (f (g x)))) ; B
(define (curry f) (lambda (a1) (lambda (a2) (f a1 a2))))
(define (apply-to a) (lambda (f) (f a))) ; C*
(define (flip f) (lambda (x y) (f y x))) ; C
(define (constantly x) (lambda args x)) ; K
(define (protect pred fn val) (if (pred val) (fn val) val))

; generic ones

(define (pipeline f . rest)
  (if (null? rest) f
    (let ((restfn (apply pipeline rest)))
      (lambda args
	(call-with-values (lambda () (apply f args)) restfn)))))

(define (bindn f . args1)
  (lambda args2 (apply f (append args1 args2))))

(define (apply-with . args) (lambda (f) (apply f args)))

(define (dispatch-collect connective . fns)
  (lambda (val)
    (apply connective (map (apply-to val) fns))))

(define (project-arg n) (lambda args (list-ref args n)))

;; predicate handling: type combinators

(define (allof . preds)
  (cond ((null? preds) (lambda (val) #t))
	((null? (cdr preds)) (car preds))
	(else
	  (let ((rest (apply allof (cdr preds))))
	    (lambda (val) (and ((car preds) val) (rest val)))))))

(define (anyof . preds)
  (cond ((null? preds) (lambda (val) #f))
	((null? (cdr preds)) (car preds))
	(else
	  (let ((rest (apply anyof (cdr preds))))
	    (lambda (val) (or ((car preds) val) (rest val)))))))

(define (compl pred) (lambda (val) (not (pred val))))

(define (member?? set) (lambda (val) (member val set)))
(define (equal?? item) (lambda (val) (equal? val item)))

